"""Callbacks which can be subscribed to by the Bluesky RunEngine in order to perform
external interactions in response to the 'documents' emitted when events occur in the
execution of an experimental plan.

Callbacks used for the Hyperion fast grid scan are prefixed with 'FGS'.
"""
